// This is the part of MPFRCPP, the C++ interface for MPFR library.

// Copyright (C) 2006 Alexey Beshenov <bav.272304@gmail.com>

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
// MA 02110-1301 USA.

// I am the <mpfrcpp/mpfrcpp.hpp> file. Include me to use MPFRCPP.

#ifndef __MPFRCPP_HPP
#define __MPFRCPP_HPP

#define MPFR_VERSION 0.005

#include <mpfr.h>
#include <gmp.h>
#include <gmpxx.h>

#ifdef __GMP_PLUSPLUS__
#define GMP_CPP_INTERFACE
#endif

#include <string>
#include <iostream>
#include <sstream>
#include <limits>

#include "config.hpp"

class MPFR
    {

    protected:

// Number:
        mpfr_t mpc_x;

// Rounding mode _for_number_. It is not MPFR item, there
// are no `rounding mode _for_number_' in MPFR.
// AVOID DIRECT ACCESS TO mpc_round_mode AND USE
// std::float_round_style TYPE
        mp_rnd_t mpc_round_mode;

// Number base. mpc_base is needed for input/output ONLY
        int mpc_base;

//------------------------------------------------------------

        static inline mp_rnd_t frs2mprnd(const std::float_round_style);
        static inline std::float_round_style mprnd2frs(const mp_rnd_t);

//------------------------------------------------------------

    public:

// We should include `error.hpp' IMMEDIATELY HERE
    #include "error.hpp"

        void base(const int b) {
            if (b >= 2 && b <= 36) { mpc_base = b; }
            else { throw error(__FILE__, __LINE__, error::invalid_base); }
            }
        int base () const { return mpc_base; }

        void round_mode(const std::float_round_style rnd) { mpc_round_mode = frs2mprnd(rnd); }
        std::float_round_style round_mode () const { return mprnd2frs(mpc_round_mode); }

        void prec(const mp_prec_t prec) { mpfr_set_prec(mpc_x, prec); }
        mp_prec_t prec () const { return mpfr_get_prec(mpc_x); }

        void exp(const mp_exp_t exp) {
            if (exp <= MPFR::emax() && exp <= MPFR::emin())
                mpfr_set_exp(mpc_x, exp);
            }
        mp_exp_t exp () const { return mpfr_get_exp(mpc_x); }

        void subnormalize(const int t) { mpfr_subnormalize (mpc_x, t, mpc_round_mode); }

        template<typename T> bool fits () const;

        int check_range () {
            int t;
            mpfr_check_range(mpc_x, t, mpc_round_mode);
            return t;
            }

//------------------------------------------------------------

// WARNING! `static's will change defaults for ALL
// INITIALIZATIONS!

// Usage: `MPFR::static()'

        inline static void default_prec(const mp_prec_t prec) { DEFAULT_PREC = prec; }
        inline static mp_prec_t default_prec() { return DEFAULT_PREC; }

        inline static void default_base(const int b) {
            if (b >= 2 && b <= 36) { DEFAULT_BASE = b; }
            else { throw error(__FILE__, __LINE__, error::invalid_base); }
            }
        inline static int default_base() { return DEFAULT_BASE; }

        inline static void set_inintd_to_0(const bool s) {
            SET_INITD_TO_0 = s;
            }
        inline static bool set_inintd_to_0() { return SET_INITD_TO_0; }

        inline static void default_round_mode(const std::float_round_style rnd) { DEFAULT_ROUND_MODE = rnd; }
        inline static std::float_round_style default_round_mode() { return DEFAULT_ROUND_MODE; }

        inline static int emin(const mp_exp_t exp) { return mpfr_set_emin(exp); }
        inline static mp_exp_t emin() { return mpfr_get_emin(); }

        inline static int emax(const mp_exp_t exp) { return mpfr_set_emax(exp); }
        inline static mp_exp_t emax() { return mpfr_get_emax(); }

        inline static mp_exp_t emin_min() { return mpfr_get_emin_min(); }
        inline static mp_exp_t emin_max() { return mpfr_get_emin_max(); }
        inline static mp_exp_t emax_min() { return mpfr_get_emax_min(); }
        inline static mp_exp_t emax_max() { return mpfr_get_emax_max(); }

//------------------------------------------------------------

// Cache for constants
        class cache
            {
            public:
                inline static void free () { return mpfr_free_cache (); }
            };

// Global flags
        class flags
            {
            public:

                inline static void clear () {
                    mpfr_clear_flags();
                    }

                inline static bool underflow () {
                    if (mpfr_underflow_p() == 0) { return false; }
                    else { return true; }
                    }
                inline static bool overflow () {
                    if (mpfr_overflow_p() == 0) { return false; }
                    else { return true; }
                    }
                inline static bool nan () {
                    if (mpfr_nanflag_p() == 0) { return false; }
                    else { return true; }
                    }
                inline static bool inexact () {
                    if (mpfr_inexflag_p() == 0) { return false; }
                    else { return true; }
                    }
                inline static bool erange () {
                    if (mpfr_erangeflag_p() == 0) { return false; }
                    else { return true; }
                    }

                inline static void underflow (const bool a) {
                    if (a == true) { mpfr_set_underflow(); }
                    if (a == false) { mpfr_clear_underflow(); }
                    }
                inline static void overflow (const bool a) {
                    if (a == true) { mpfr_set_overflow(); }
                    if (a == false) { mpfr_clear_overflow(); }
                    }
                inline static void nan (const bool a) {
                    if (a == true) { mpfr_set_nanflag(); }
                    if (a == false) { mpfr_clear_nanflag(); }
                    }
                inline static void inexact (const bool a) {
                    if (a == true) { mpfr_set_inexflag(); }
                    if (a == false) { mpfr_clear_inexflag(); }
                    }
                inline static void erange (const bool a) {
                    if (a == true) { mpfr_set_erangeflag(); }
                    if (a == false) { mpfr_clear_erangeflag(); }
                    }

            };

//------------------------------------------------------------

// Constructors:

// We should include `constructors.cpp' IMMEDIATELY HERE
#include "constructors.cpp"

//------------------------------------------------------------

// Special assignments:

        void inf (const int sign) { mpfr_set_inf(mpc_x, sign); }
        void nan () { mpfr_set_nan(mpc_x); }

//------------------------------------------------------------

// Assignment/conversion functions

        void operator= (const MPFR&);
        void operator= (const unsigned long int);
        void operator= (const unsigned int);
        void operator= (const long int);
        void operator= (const int);
        void operator= (const double&);
        void operator= (const long double&);
        void operator= (const mpz_t&);
        void operator= (const mpq_t&);
        void operator= (const mpf_t&);
#ifdef GMP_CPP_INTERFACE
        void operator= (const mpz_class& y);
        void operator= (const mpq_class& y);
        void operator= (const mpf_class& y);
#endif

        void operator= (const std::string&);
        void operator= (const char *s);
        void operator= (const bool y);

        operator unsigned long int() const;
        operator unsigned int() const;
        operator long int() const;
        operator int() const;
        operator double () const;
        operator long double () const;
#ifdef GMP_CPP_INTERFACE
        operator mpz_class () const;
        operator mpf_class () const;
#endif
        operator std::string () const;

//------------------------------------------------------------

// Rounding

        void round (const mp_prec_t, const std::float_round_style);

//------------------------------------------------------------

// Basic Arithmetic Functions

        friend MPFR abs (const MPFR &);
        friend MPFR sqrt (const MPFR &);
        friend MPFR cbrt (const MPFR &);
        friend MPFR root (const MPFR &, const unsigned long int);
        friend MPFR sqr (const MPFR &);

        friend MPFR pow (const MPFR &, const MPFR &);
        friend MPFR pow (const MPFR &, const unsigned long int &);
        friend MPFR pow (const unsigned long int &, const MPFR &);
        friend MPFR pow (const MPFR &, const long int &);
        friend MPFR pow (const long int &, const MPFR &);
        friend MPFR pow (const MPFR &, const mpz_t &);
        friend MPFR pow (const unsigned long int &, const unsigned long int &, const mp_prec_t, const std::float_round_style);
        friend MPFR pow (const unsigned long int &, const unsigned int &, const mp_prec_t, const std::float_round_style);

        friend MPFR operator+ (const MPFR &, const MPFR &);
        friend MPFR operator+ (const MPFR &, const unsigned long int &);
        friend MPFR operator+ (const MPFR &, const long int &);
        friend MPFR operator+ (const MPFR &, const mpz_t &);
        friend MPFR operator+ (const MPFR &, const mpq_t &);

        friend MPFR operator- (const MPFR &, const MPFR &);
        friend MPFR operator- (const MPFR &, const unsigned long int &);
        friend MPFR operator- (const unsigned long int &, const MPFR &);
        friend MPFR operator- (const MPFR &, const long int &);
        friend MPFR operator- (const long int &, const MPFR &);
        friend MPFR operator- (const MPFR &, const mpz_t &);
        friend MPFR operator- (const MPFR &, const mpq_t &);

        MPFR operator- () const;

        friend MPFR operator* (const MPFR &, const MPFR &);
        friend MPFR operator* (const MPFR &, const unsigned long int &);
        friend MPFR operator* (const MPFR &, const long int &);
        friend MPFR operator* (const MPFR &, const mpz_t &);
        friend MPFR operator* (const MPFR &, const mpq_t &);

        friend MPFR operator/ (const MPFR &, const MPFR &);
        friend MPFR operator/ (const MPFR &, const unsigned long int &);
        friend MPFR operator/ (const unsigned long int &, const MPFR &);
        friend MPFR operator/ (const MPFR &, const long int &);
        friend MPFR operator/ (const long int &, const MPFR &);
        friend MPFR operator/ (const MPFR &, const mpz_t &);
        friend MPFR operator/ (const MPFR &, const mpq_t &);

        MPFR operator-- (int);
        MPFR operator++ (int);
        MPFR operator-- ();
        MPFR operator++ ();

//------------------------------------------------------------

// Special Functions

        friend MPFR exp (const MPFR &);
        friend MPFR expm1 (const MPFR &);
        friend MPFR exp10 (const MPFR &);
        friend MPFR exp2 (const MPFR &);

        friend MPFR log (const MPFR &);
        friend MPFR log10 (const MPFR &);
        friend MPFR log2 (const MPFR &);
        friend MPFR log1p (const MPFR &);

        friend MPFR sin (const MPFR &);
        friend MPFR cos (const MPFR &);
        friend MPFR sec (const MPFR &);
        friend MPFR csc (const MPFR &);
        friend MPFR tan (const MPFR &);
        friend MPFR cot (const MPFR &);

        friend MPFR asin (const MPFR &);
        friend MPFR acos (const MPFR &);
        friend MPFR atan (const MPFR &);
        friend MPFR atan2 (const MPFR &, const MPFR &);

        friend MPFR sinh (const MPFR &);
        friend MPFR cosh (const MPFR &);
        friend MPFR sech (const MPFR &);
        friend MPFR csch (const MPFR &);
        friend MPFR tanh (const MPFR &);
        friend MPFR coth (const MPFR &);

        friend MPFR asinh (const MPFR &);
        friend MPFR acosh (const MPFR &);
        friend MPFR atanh (const MPFR &);

        friend MPFR erf (const MPFR &);
        friend MPFR erfc (const MPFR &);
        friend MPFR eint (const MPFR &);
        friend MPFR gamma (const MPFR &);
        friend MPFR lngamma (const MPFR &);
        friend MPFR zeta (const MPFR &);

        friend MPFR agm (const MPFR &, const MPFR &);
        friend MPFR hypot (const MPFR &, const MPFR &);
        friend MPFR factorial (const unsigned long int &, mp_prec_t, std::float_round_style);

        friend MPFR const_catalan (const mp_prec_t, const std::float_round_style, const bool);
        friend MPFR const_euler (const mp_prec_t, const std::float_round_style, const bool);
        friend MPFR const_log2 (const mp_prec_t, const std::float_round_style, const bool);
        friend MPFR const_pi (const mp_prec_t, const std::float_round_style, const bool);

        friend MPFR random2 (const mp_size_t, const mp_exp_t, const mp_prec_t, const std::float_round_style);
        friend MPFR urandomb (gmp_randstate_t, const mp_prec_t, const std::float_round_style);

//------------------------------------------------------------

// Comparison functions

        friend bool is_nan (const MPFR&);
        friend bool is_inf (const MPFR&);
        friend bool is_number (const MPFR&);
        friend bool is_zero (const MPFR&);
        friend bool is_unordered (const MPFR&, const MPFR&);
        friend int sgn (const MPFR&);
        friend bool operator> (const MPFR&, const MPFR&);
        friend bool operator<= (const MPFR&, const MPFR&);
        friend bool operator< (const MPFR&, const MPFR&);
        friend bool operator>= (const MPFR&, const MPFR&);
        friend bool operator== (const MPFR&, const MPFR&);

//------------------------------------------------------------

// Integer related functions

        friend bool is_integer (const MPFR&);

        friend MPFR ceil (const MPFR &);
        friend MPFR floor (const MPFR &);
        friend MPFR round (const MPFR &);
        friend MPFR trunc (const MPFR &);
        friend MPFR frac (const MPFR &);

//------------------------------------------------------------

// Stream

        template<typename CharT, class Traits> friend std::basic_ostream<CharT, Traits>& operator << (std::basic_ostream<CharT, Traits>&, const MPFR&);
        template<typename CharT, class Traits> friend int operator>> (std::basic_istream<CharT, Traits>&, MPFR&);

//------------------------------------------------------------

// Some useful MPFR functions without MPFRCPP interface

        friend int mpfr_set_ui_2exp (mpfr_t, unsigned long int, mp_exp_t, mp_rnd_t);
        friend int mpfr_set_si_2exp (mpfr_t, long int, mp_exp_t, mp_rnd_t);

        /*
        friend int mpfr_set_uj_2exp (mpfr_t, uintmax_t, intmax_t, mp_rnd_t);
        friend int mpfr_set_sj_2exp (mpfr_t, intmax_t, intmax_t, mp_rnd_t);
void mpfr_swap (mpfr_t, mpfr_t);
        */

        friend int mpfr_mul_2ui (mpfr_t, mpfr_t, unsigned long int, mp_rnd_t);
        friend int mpfr_mul_2si (mpfr_t, mpfr_t, long int, mp_rnd_t);
        friend int mpfr_div_2ui (mpfr_t, mpfr_t, unsigned long int, mp_rnd_t);
        friend int mpfr_div_2si (mpfr_t, mpfr_t, long int, mp_rnd_t);

        /*
        friend int mpfr_cmp (mpfr_t, mpfr_t);
        friend int mpfr_cmp_ui (mpfr_t, unsigned long int);
        friend int mpfr_cmp_si (mpfr_t, long int);
        */

        friend int mpfr_cmp_d (mpfr_t, double);
        friend int mpfr_cmp_ld (mpfr_t, long double);
        friend int mpfr_cmp_z (mpfr_t, mpz_t);
        friend int mpfr_cmp_q (mpfr_t, mpq_t);
        friend int mpfr_cmp_f (mpfr_t, mpf_t);
        friend int mpfr_cmp_ui_2exp (mpfr_t, unsigned long int, mp_exp_t);
        friend int mpfr_cmp_si_2exp (mpfr_t, long int, mp_exp_t);
        friend int mpfr_cmpabs (mpfr_t, mpfr_t);
        friend int mpfr_sin_cos (mpfr_t, mpfr_t, mpfr_t, mp_rnd_t);

    };

//------------------------------------------------------------

// This *.cpp's contain functions and operators:

#include "float_round_style.cpp"
#include "assignment_conversion.cpp"
#include "comparison.cpp"
#include "fits.cpp"

#include "rounding.cpp"
#include "basic_arithmetic.cpp"
#include "special.cpp"
#include "integer_related.cpp"
#include "stream.cpp"

#endif                                            // __MPFRCPP_HPP
